/*
 * Created on 01-jul-2005
 *
 * gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 *
 * Copyright (C) 2004 IVER T.I. and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *  Generalitat Valenciana
 *   Conselleria d'Infraestructures i Transport
 *   Av. Blasco Ibez, 50
 *   46010 VALENCIA
 *   SPAIN
 *
 *      +34 963862235
 *   gvsig@gva.es
 *      www.gvsig.gva.es
 *
 *    or
 *
 *   IVER T.I. S.A
 *   Salamanca 50
 *   46005 Valencia
 *   Spain
 *
 *   +34 963163400
 *   dac@iver.es
 */
package com.iver.gvsig.geoprocessing.gui;

import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JTextField;
import javax.swing.JScrollPane;

import com.hardcode.driverManager.DriverLoadException;
import com.hardcode.gdbms.engine.data.DataSource;
import com.hardcode.gdbms.engine.data.DataSourceFactory;
import com.hardcode.gdbms.engine.data.NoSuchTableException;
import com.hardcode.gdbms.engine.data.edition.DataWare;
import com.hardcode.gdbms.engine.instruction.IncompatibleTypesException;
import com.hardcode.gdbms.engine.values.BooleanValue;
import com.hardcode.gdbms.engine.values.Value;
import com.hardcode.gdbms.engine.values.ValueFactory;
import com.iver.andami.PluginServices;
import com.iver.andami.ui.mdiManager.View;
import com.iver.andami.ui.mdiManager.ViewInfo;
import com.iver.cit.gvsig.fmap.DriverException;
import com.iver.cit.gvsig.fmap.core.FPolygon2D;
import com.iver.cit.gvsig.fmap.core.FShape;
import com.iver.cit.gvsig.fmap.core.IGeometry;
import com.iver.cit.gvsig.fmap.core.ShapeFactory;
import com.iver.cit.gvsig.fmap.core.v02.FConverter;
import com.iver.cit.gvsig.fmap.drivers.DriverIOException;
import com.iver.cit.gvsig.fmap.layers.FBitSet;
import com.iver.cit.gvsig.fmap.layers.FLayer;
import com.iver.cit.gvsig.fmap.layers.FLayers;
import com.iver.cit.gvsig.fmap.layers.SelectableDataSource;
import com.iver.cit.gvsig.fmap.layers.VectorialAdapter;
import com.iver.cit.gvsig.fmap.layers.layerOperations.AlphanumericData;
import com.iver.cit.gvsig.fmap.layers.layerOperations.Selectable;
import com.iver.cit.gvsig.fmap.layers.layerOperations.SingleLayer;
import com.iver.cit.gvsig.fmap.layers.layerOperations.VectorialData;
import com.iver.cit.gvsig.fmap.operations.strategies.SelectedShapeVisitor;
import com.iver.cit.gvsig.fmap.operations.strategies.SelectedZoomVisitor;
import com.iver.cit.gvsig.fmap.operations.strategies.VisitException;
import com.iver.gvsig.geoprocessing.gui.operationpanels.GeoProcessingBufferPanel;
import com.iver.gvsig.geoprocessing.gui.operationpanels.GeoProcessingClipPanel;
import com.iver.gvsig.geoprocessing.gui.operationpanels.GeoProcessingDissolvePanel;
import com.iver.gvsig.geoprocessing.gui.operationpanels.GeoProcessingIntersectPanel;
import com.iver.gvsig.geoprocessing.gui.operationpanels.GeoProcessingMergePanel;
import com.iver.gvsig.geoprocessing.gui.operationpanels.GeoProcessingSpatialjoinPanel;
import com.iver.gvsig.geoprocessing.gui.operationpanels.GeoProcessingUnionPanel;
import com.iver.gvsig.geoprocessing.operations.GeoProcessingPersistenceOperations;
import com.iver.gvsig.geoprocessing.operations.GeoProcessingTableOperations;
import com.iver.gvsig.geoprocessing.operations.GeoProcessingTopologyOperations;
import com.iver.gvsig.geoprocessing.operations.strategies.SelectedShapeToJTSVisitor;
import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryCollection;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.MultiPoint;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.Polygon;
import com.vividsolutions.jts.operation.distance.DistanceOp;

import java.awt.Component;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.FlowLayout;
import java.awt.CardLayout;
import javax.swing.BoxLayout;
import java.awt.BorderLayout;
import java.io.File;
import java.sql.Types;
import java.util.BitSet;
import java.util.Vector;

import javax.swing.JButton;

public class GeoProcessingPanel extends JPanel implements View {

    /**
     * 
     */
    private static final long serialVersionUID = 1L;
    private GeoProcessingOperationSelectorPanel geoProcessingOperationSelectorPanel = null;
    private GeoProcessingBufferPanel geoProcessingBufferPanel = null;
    private GeoProcessingClipPanel geoProcessingClipPanel = null;
    private GeoProcessingDissolvePanel geoProcessingDissolvePanel = null;
    private GeoProcessingMergePanel geoProcessingMergePanel = null;
    private GeoProcessingIntersectPanel geoProcessingIntersectPanel = null;
    private GeoProcessingUnionPanel geoProcessingUnionPanel = null;
    private GeoProcessingSpatialjoinPanel geoProcessingSpatialjoinPanel = null;
    private JPanel buttonsPanel = null;

	private JButton jButton = null;
	private JButton jButton1 = null;
	private JButton jButton2 = null;
    private ViewInfo viewInfo = null;
    private FLayers layers = null;
    /**
     * This constructor initializes the set of layers
     */
    public GeoProcessingPanel(FLayers layers) {
        super();
        this.layers = layers;
        initialize();
    }

    /**
     * This method initializes this
     * 
     * @return void
     */
    private void initialize() {
        this.setLayout(new BorderLayout());
        this.setSize(416, 271);
        this.add(getButtonsPanel(), java.awt.BorderLayout.SOUTH);
        this.add(getGeoProcessingOperationSelectorPanel(), java.awt.BorderLayout.NORTH);
        this.add(getGeoProcessingBufferPanel(), java.awt.BorderLayout.NORTH);
        this.add(getGeoProcessingClipPanel(), java.awt.BorderLayout.NORTH);
        this.add(getGeoProcessingDissolvePanel(), java.awt.BorderLayout.NORTH);
        this.add(getGeoProcessingMergePanel(), java.awt.BorderLayout.NORTH);
        this.add(getGeoProcessingIntersectPanel(), java.awt.BorderLayout.NORTH);
        this.add(getGeoProcessingUnionPanel(), java.awt.BorderLayout.NORTH);
        this.add(getGeoProcessingSpatialjoinPanel(), java.awt.BorderLayout.NORTH);
        geoProcessingOperationSelectorPanel.setVisible(true);
        geoProcessingBufferPanel.setVisible(false);
        geoProcessingClipPanel.setVisible(false);
        geoProcessingDissolvePanel.setVisible(false);
        geoProcessingMergePanel.setVisible(false);
        geoProcessingIntersectPanel.setVisible(false);
        geoProcessingUnionPanel.setVisible(false);
        geoProcessingSpatialjoinPanel.setVisible(false);
        jButton1.setEnabled(false);
    }

    public ViewInfo getViewInfo() {
        // TODO Auto-generated method stub
        if (viewInfo == null) {
            viewInfo=new ViewInfo(ViewInfo.MODALDIALOG);
            viewInfo.setTitle(PluginServices.getText(this,"Herramientas_de_analisis"));
        }
        return viewInfo;
    }

    /**
     * This method initializes geoProcessingOperationSelectorPanel	
     * 	
     * @return javax.swing.JPanel	
     */    
    private JPanel getGeoProcessingOperationSelectorPanel() {
    	if (geoProcessingOperationSelectorPanel == null) {
    		geoProcessingOperationSelectorPanel = new GeoProcessingOperationSelectorPanel();
    		geoProcessingOperationSelectorPanel.setName("geoProcessingOperationSelectorPanel");
    	}
    	return geoProcessingOperationSelectorPanel;
    }

    /**
     * This method initializes geoProcessingBufferPanel	
     * 	
     * @return javax.swing.JPanel	
     */    
    private JPanel getGeoProcessingBufferPanel() {
    	if (geoProcessingBufferPanel == null) {
    		geoProcessingBufferPanel = new GeoProcessingBufferPanel(layers);
    		geoProcessingBufferPanel.setName("geoProcessingBufferPanel");
    	}
    	return geoProcessingBufferPanel;
    }

    /**
     * This method initializes geoProcessingClipPanel	
     * 	
     * @return javax.swing.JPanel	
     */    
    private JPanel getGeoProcessingClipPanel() {
    	if (geoProcessingClipPanel == null) {
    		geoProcessingClipPanel = new GeoProcessingClipPanel(layers);
    		geoProcessingClipPanel.setName("geoProcessingClipPanel");
    		// Si no le meto esta lnea, no se visualiza el men. Ver que puede
            // estar pasando ...
            geoProcessingClipPanel.setPreferredSize(new java.awt.Dimension(300,300));
    	}
    	return geoProcessingClipPanel;
    }

    /**
     * This method initializes geoProcessingDissolvePanel 
     *  
     * @return javax.swing.JPanel   
     */    
    private JPanel getGeoProcessingDissolvePanel() {
        if (geoProcessingDissolvePanel == null) {
            geoProcessingDissolvePanel = new GeoProcessingDissolvePanel(layers);
            geoProcessingDissolvePanel.setName("geoProcessingDissolvePanel");
        }
        return geoProcessingDissolvePanel;
    }

    /**
     * This method initializes geoProcessingMergePanel 
     *  
     * @return javax.swing.JPanel   
     */    
    private JPanel getGeoProcessingMergePanel() {
        if (geoProcessingMergePanel == null) {
            geoProcessingMergePanel = new GeoProcessingMergePanel(layers);
            geoProcessingMergePanel.setName("geoProcessingMergePanel");
        }
        return geoProcessingMergePanel;
    }

    /**
     * This method initializes geoProcessingIntersectPanel 
     *  
     * @return javax.swing.JPanel   
     */    
    private JPanel getGeoProcessingIntersectPanel() {
        if (geoProcessingIntersectPanel == null) {
            geoProcessingIntersectPanel = new GeoProcessingIntersectPanel(layers);
            geoProcessingIntersectPanel.setName("geoProcessingIntersectPanel");
        }
        return geoProcessingIntersectPanel;
    }

    /**
     * This method initializes geoProcessingUnionPanel 
     *  
     * @return javax.swing.JPanel   
     */    
    private JPanel getGeoProcessingUnionPanel() {
        if (geoProcessingUnionPanel == null) {
            geoProcessingUnionPanel = new GeoProcessingUnionPanel(layers);
            geoProcessingUnionPanel.setName("geoProcessingUnionPanel");
        }
        return geoProcessingUnionPanel;
    }

    /**
     * This method initializes geoProcessingSpatialjoinPanel 
     *  
     * @return javax.swing.JPanel   
     */    
    private JPanel getGeoProcessingSpatialjoinPanel() {
        if (geoProcessingSpatialjoinPanel == null) {
            geoProcessingSpatialjoinPanel = new GeoProcessingSpatialjoinPanel(layers);
            geoProcessingSpatialjoinPanel.setName("geoProcessingSpatialjoinPanel");
            // Si no le meto esta lnea, no se visualiza el men. Ver que puede
            // estar pasando ...
            geoProcessingSpatialjoinPanel.setPreferredSize(new java.awt.Dimension(300,300));
        }
        return geoProcessingSpatialjoinPanel;
    }

    /**
     * This method initializes buttonsPanel	
     * 	
     * @return javax.swing.JPanel	
     */    
    private JPanel getButtonsPanel() {
    	if (buttonsPanel == null) {
    		buttonsPanel = new JPanel();
    		buttonsPanel.setName("buttonsPanel");
    		buttonsPanel.setBorder(javax.swing.BorderFactory.createEtchedBorder(javax.swing.border.EtchedBorder.LOWERED));
    		buttonsPanel.add(getJButton(), null);
    		buttonsPanel.add(getJButton1(), null);
    		buttonsPanel.add(getJButton2(), null);
    	}
    	return buttonsPanel;
    }

	/**
	 * This method initializes jButton	
	 * 	
	 * @return javax.swing.JButton	
	 */    
	private JButton getJButton() {
		if (jButton == null) {
			jButton = new JButton();
			jButton.setText(PluginServices.getText(this,"Cerrar"));
			jButton.addActionListener(new java.awt.event.ActionListener() { 
				public void actionPerformed(java.awt.event.ActionEvent e) {    
					System.out.println("actionPerformed()"); // TODO Auto-generated Event stub actionPerformed()
                    closeMethod();
				}
			});
		}
		return jButton;
	}
	/**
	 * This method initializes jButton1	
	 * 	
	 * @return javax.swing.JButton	
	 */    
	private JButton getJButton1() {
		if (jButton1 == null) {
			jButton1 = new JButton();
			jButton1.setText(PluginServices.getText(this,"Anterior"));
			jButton1.addActionListener(new java.awt.event.ActionListener() { 
				public void actionPerformed(java.awt.event.ActionEvent e) {    
					System.out.println("actionPerformed()"); // TODO Auto-generated Event stub actionPerformed()
                    getGeoProcessingOperationSelectorPanel().setVisible(true);
                    getGeoProcessingBufferPanel().setVisible(false);
                    getGeoProcessingClipPanel().setVisible(false);
                    getGeoProcessingDissolvePanel().setVisible(false);
                    getGeoProcessingMergePanel().setVisible(false);
                    getGeoProcessingIntersectPanel().setVisible(false);
                    getGeoProcessingUnionPanel().setVisible(false);
                    getGeoProcessingSpatialjoinPanel().setVisible(false);
                    jButton1.setEnabled(false);
                    jButton2.setText(PluginServices.getText(this,"Siguiente"));
				}
			});
		}
		return jButton1;
	}
	/**
	 * This method initializes jButton2	
	 * 	
	 * @return javax.swing.JButton	
	 */    
	private JButton getJButton2() {
		if (jButton2 == null) {
			jButton2 = new JButton();
			jButton2.setText(PluginServices.getText(this,"Siguiente"));
			jButton2.addActionListener(new java.awt.event.ActionListener() { 
				public void actionPerformed(java.awt.event.ActionEvent e) {    
					System.out.println("actionPerformed()"); // TODO Auto-generated Event stub actionPerformed()
					if (jButton2.getText().equals(PluginServices.getText(this,"Siguiente"))) {
                        if (((JRadioButton)geoProcessingOperationSelectorPanel.getComponent(1)).isSelected()) {
                            getGeoProcessingOperationSelectorPanel().setVisible(false);
                            getGeoProcessingBufferPanel().setVisible(true);
                            getGeoProcessingClipPanel().setVisible(false);
                            getGeoProcessingDissolvePanel().setVisible(false);
                            getGeoProcessingMergePanel().setVisible(false);
                            getGeoProcessingIntersectPanel().setVisible(false);
                            getGeoProcessingUnionPanel().setVisible(false);
                            getGeoProcessingSpatialjoinPanel().setVisible(false);
                        } else if (((JRadioButton)geoProcessingOperationSelectorPanel.getComponent(2)).isSelected()) {
                            getGeoProcessingOperationSelectorPanel().setVisible(false);
                            getGeoProcessingBufferPanel().setVisible(false);
                            getGeoProcessingClipPanel().setVisible(true);
                            getGeoProcessingDissolvePanel().setVisible(false);
                            getGeoProcessingMergePanel().setVisible(false);
                            getGeoProcessingIntersectPanel().setVisible(false);
                            getGeoProcessingUnionPanel().setVisible(false);
                            getGeoProcessingSpatialjoinPanel().setVisible(false);
                        } else if (((JRadioButton)geoProcessingOperationSelectorPanel.getComponent(3)).isSelected()) {
                            getGeoProcessingOperationSelectorPanel().setVisible(false);
                            getGeoProcessingBufferPanel().setVisible(false);
                            getGeoProcessingClipPanel().setVisible(false);
                            getGeoProcessingDissolvePanel().setVisible(true);
                            getGeoProcessingMergePanel().setVisible(false);
                            getGeoProcessingIntersectPanel().setVisible(false);
                            getGeoProcessingUnionPanel().setVisible(false);
                            getGeoProcessingSpatialjoinPanel().setVisible(false);
                        } else if (((JRadioButton)geoProcessingOperationSelectorPanel.getComponent(4)).isSelected()) {
                            getGeoProcessingOperationSelectorPanel().setVisible(false);
                            getGeoProcessingBufferPanel().setVisible(false);
                            getGeoProcessingClipPanel().setVisible(false);
                            getGeoProcessingDissolvePanel().setVisible(false);
                            getGeoProcessingMergePanel().setVisible(true);
                            getGeoProcessingIntersectPanel().setVisible(false);
                            getGeoProcessingUnionPanel().setVisible(false);
                            getGeoProcessingSpatialjoinPanel().setVisible(false);
                        } else if (((JRadioButton)geoProcessingOperationSelectorPanel.getComponent(5)).isSelected()) {
                            getGeoProcessingOperationSelectorPanel().setVisible(false);
                            getGeoProcessingBufferPanel().setVisible(false);
                            getGeoProcessingClipPanel().setVisible(false);
                            getGeoProcessingDissolvePanel().setVisible(false);
                            getGeoProcessingMergePanel().setVisible(false);
                            getGeoProcessingIntersectPanel().setVisible(true);
                            getGeoProcessingUnionPanel().setVisible(false);
                            getGeoProcessingSpatialjoinPanel().setVisible(false);
                        } else if (((JRadioButton)geoProcessingOperationSelectorPanel.getComponent(6)).isSelected()) {
                            getGeoProcessingOperationSelectorPanel().setVisible(false);
                            getGeoProcessingBufferPanel().setVisible(false);
                            getGeoProcessingClipPanel().setVisible(false);
                            getGeoProcessingDissolvePanel().setVisible(false);
                            getGeoProcessingMergePanel().setVisible(false);
                            getGeoProcessingIntersectPanel().setVisible(false);
                            getGeoProcessingUnionPanel().setVisible(true);
                            getGeoProcessingSpatialjoinPanel().setVisible(false);
                        } else if (((JRadioButton)geoProcessingOperationSelectorPanel.getComponent(7)).isSelected()) {
                            getGeoProcessingOperationSelectorPanel().setVisible(false);
                            getGeoProcessingBufferPanel().setVisible(false);
                            getGeoProcessingClipPanel().setVisible(false);
                            getGeoProcessingDissolvePanel().setVisible(false);
                            getGeoProcessingMergePanel().setVisible(false);
                            getGeoProcessingIntersectPanel().setVisible(false);
                            getGeoProcessingUnionPanel().setVisible(false);
                            getGeoProcessingSpatialjoinPanel().setVisible(true);
                        } else {
                            // Si no hay ninguna operacin de geoprocesamiento seleccionada no se cambia el dilogo de opciones.
                            getGeoProcessingOperationSelectorPanel().setVisible(true);
                            getGeoProcessingBufferPanel().setVisible(false);
                            getGeoProcessingClipPanel().setVisible(false);
                            getGeoProcessingDissolvePanel().setVisible(false);
                            getGeoProcessingMergePanel().setVisible(false);
                            getGeoProcessingIntersectPanel().setVisible(false);
                            getGeoProcessingUnionPanel().setVisible(false);
                            getGeoProcessingSpatialjoinPanel().setVisible(false);
                        }
                        jButton1.setEnabled(true);
                        jButton2.setText(PluginServices.getText(this,"Terminar"));
                    } else if (jButton2.getText().equals(PluginServices.getText(this,"Terminar"))) {
                        if (((JRadioButton)geoProcessingOperationSelectorPanel.getComponent(1)).isSelected()) {
                            bufferMethod();
                        } else if (((JRadioButton)geoProcessingOperationSelectorPanel.getComponent(2)).isSelected()) {
                            clipMethod();
                        } else if (((JRadioButton)geoProcessingOperationSelectorPanel.getComponent(3)).isSelected()) {
                            dissolveMethod();
                        } else if (((JRadioButton)geoProcessingOperationSelectorPanel.getComponent(4)).isSelected()) {
                            mergeMethod();
                        } else if (((JRadioButton)geoProcessingOperationSelectorPanel.getComponent(5)).isSelected()) {
                            intersectMethod();
                        } else if (((JRadioButton)geoProcessingOperationSelectorPanel.getComponent(6)).isSelected()) {
                            unionMethod();
                        } else if (((JRadioButton)geoProcessingOperationSelectorPanel.getComponent(7)).isSelected()) {
                            spatialjoinMethod();
                        } else {
                            //
                        }
                        closeMethod();
                    }
				}
			});
		}
		return jButton2;
	}
    private Geometry[] getJtsGeometries(FLayer bufferingLayer, boolean selected) {
        FBitSet fBitSet = ((Selectable)bufferingLayer).getSelection();
        // Extraigo todos los elementos de la capa o solo los seleccionados
        //if (!((JCheckBox)component).isSelected()) {
        if (!selected) {
            VectorialAdapter va = (VectorialAdapter)((SingleLayer)bufferingLayer).getSource();
            fBitSet = new FBitSet();
            try {
                for (int i=0;i<va.getShapeCount();i++) {
                   fBitSet.set(i);
                }
            } catch (DriverIOException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
            ((Selectable)bufferingLayer).setSelection(fBitSet);
        }
        SelectedShapeToJTSVisitor visitor = new SelectedShapeToJTSVisitor();
        try {
            ((VectorialData)bufferingLayer).process(visitor);
        } catch (DriverException e1) {
            // TODO Auto-generated catch block
            e1.printStackTrace();
        } catch (VisitException e1) {
            // TODO Auto-generated catch block
            e1.printStackTrace();
        }
        // Si no haban elementos seleccionados, cambio la seleccin temporal
        // completa actual por la original seleccin nula
        if (!selected) ((Selectable)bufferingLayer).clearSelection();
        // Almaceno en un vector los elementos para el buffer
        return visitor.getJtsGeometries();
    }
    private IGeometry[] getFmapGeometries(Geometry[] jtsGeometries) {
        FShape[] shapes = new FShape[jtsGeometries.length];
        IGeometry[] geoms = new IGeometry[jtsGeometries.length];
        for (int i=0;i<jtsGeometries.length;i++) {
            if (jtsGeometries[i] instanceof MultiPoint) {
                double[] xcoords = new double[jtsGeometries[i].getNumGeometries()];
                double[] ycoords = new double[jtsGeometries[i].getNumGeometries()];
                for (int j=0;j<jtsGeometries[i].getNumGeometries();j++) {
                    xcoords[j] = ((Point)jtsGeometries[i].getGeometryN(j)).getX();
                    ycoords[j] = ((Point)jtsGeometries[i].getGeometryN(j)).getY();
                }
                geoms[i] = ShapeFactory.createMultipoint2D(xcoords, ycoords);
            } else {
                shapes[i] = FConverter.jts_to_java2d(jtsGeometries[i]);
                geoms[i] = ShapeFactory.createGeometry(shapes[i]);
            }
        }
        return geoms;
    }
    private void bufferMethod() {
        String bufferingLayerName = (String)((JComboBox)geoProcessingBufferPanel.getComponent(2)).getSelectedItem();
        // TODO: Provisional. Hasta que Fernando le meta PKs al resto de ficheros.
        //System.out.println("bufferingLayerName = " + bufferingLayerName);
        boolean isDbfFile = false;
        if (bufferingLayerName.endsWith(".shp") || bufferingLayerName.endsWith(".SHP")) isDbfFile = true;
        FLayer bufferingLayer = layers.getLayer(bufferingLayerName);
        boolean selected = false;
        if (((JCheckBox)geoProcessingBufferPanel.getComponent(3)).isSelected()) selected = true;
        Geometry[] geometries = getJtsGeometries(bufferingLayer, selected);
        IGeometry[] shapes = null;
        boolean dissolveBuffer = false;
        // Realizo el buffer llamando al metodo makeBuffer de GeoProcessingTopologyOperations
        if (((JRadioButton)geoProcessingBufferPanel.getComponent(4)).isSelected()) {
            //dissolveBuffer = true;
            double bufferDistance = Double.parseDouble(((JTextField)geoProcessingBufferPanel.getComponent(7)).getText());
            if (((JCheckBox)geoProcessingBufferPanel.getComponent(13)).isSelected()) {
                dissolveBuffer = true;
                Geometry jtsBuffer = GeoProcessingTopologyOperations.makeDissolveBuffer(geometries, bufferDistance, getQuadrantSegments());
                // Realizo las operaciones relacionadas con tablas que el buffer requiere llamando a los mtodos convenientes de GeoProcessingTableOperations
                // (A lo mejor este paso no es necesario) Convierto las tablas modificadas y las nuevas entidades JTS al modelo de gvSIG
                System.out.println("jtsBuffer = " + jtsBuffer);
                FPolygon2D shapeBuffer = (FPolygon2D)FConverter.jts_to_java2d(jtsBuffer);
                System.out.println("shapeBuffer = " + shapeBuffer);
                shapes = new IGeometry[]{ShapeFactory.createGeometry(shapeBuffer)};
                // Guardo la tabla mediante gdbms. Guardo tambin las entidades JTS mediante gvSIG o usando solo gdbms (probablemente bastar con usar gdbms)
                // Con dissolve. Habr que hacer una tabla para un nico polgono o
                // multipolgono. En la tabla solo meteremos los campos para los
                // que todos los elementos tengan atributos comunes. Si no hay
                // ningn campo con estas caractersticas crearemos la tabla en
                // blanco.
            } else {
                double[] bufferDistances = new double[geometries.length];
                for (int i=0;i<bufferDistances.length;i++) {
                    bufferDistances[i] = bufferDistance;
                }
                Geometry[] jtsBuffers = GeoProcessingTopologyOperations.makeBuffers(geometries, bufferDistances, getQuadrantSegments());
                // Realizo las operaciones relacionadas con tablas que el buffer requiere llamando a los mtodos convenientes de GeoProcessingTableOperations
                // (A lo mejor este paso no es necesario) Convierto las tablas modificadas y las nuevas entidades JTS al modelo de gvSIG
                shapes = getFmapGeometries(jtsBuffers);
                // Guardo la tabla mediante gdbms. Guardo tambin las entidades JTS mediante gvSIG o usando solo gdbms (probablemente bastar con usar gdbms)
                // En este caso se guarda una tabla semejante a la del tema de entrada
                // solo que ahora los atributos se corresponden con los buffers
                // producidos por las entidades a las que correspondan estos
                // atributos.
            }
        } else if (((JRadioButton)geoProcessingBufferPanel.getComponent(5)).isSelected()) {
            //dissolveBuffer = false;
            double[] bufferDistances = getBufferDistancesFromTable(bufferingLayer);
            if (((JCheckBox)geoProcessingBufferPanel.getComponent(13)).isSelected()) {
                dissolveBuffer = true;
                // Aqu hay problemas ... Con distintas distancias y haciendo dissolve ...
                Geometry jtsBuffer = GeoProcessingTopologyOperations.makeDissolveBuffer(geometries, bufferDistances, getQuadrantSegments());
                // Realizo las operaciones relacionadas con tablas que el buffer requiere llamando a los mtodos convenientes de GeoProcessingTableOperations
                // (A lo mejor este paso no es necesario) Convierto las tablas modificadas y las nuevas entidades JTS al modelo de gvSIG
                System.out.println("jtsBuffer = " + jtsBuffer);
                FPolygon2D shapeBuffer = (FPolygon2D)FConverter.jts_to_java2d(jtsBuffer);
                System.out.println("shapeBuffer = " + shapeBuffer);
                shapes = new IGeometry[]{ShapeFactory.createGeometry(shapeBuffer)};
                // Guardo la tabla mediante gdbms. Guardo tambin las entidades JTS mediante gvSIG o usando solo gdbms (probablemente bastar con usar gdbms)
            } else {
                Geometry[] jtsBuffers = GeoProcessingTopologyOperations.makeBuffers(geometries, bufferDistances, getQuadrantSegments());
                // Realizo las operaciones relacionadas con tablas que el buffer requiere llamando a los mtodos convenientes de GeoProcessingTableOperations
                // (A lo mejor este paso no es necesario) Convierto las tablas modificadas y las nuevas entidades JTS al modelo de gvSIG
                shapes = getFmapGeometries(jtsBuffers);
                // Guardo la tabla mediante gdbms. Guardo tambin las entidades JTS mediante gvSIG o usando solo gdbms (probablemente bastar con usar gdbms)
            }
        }
        AlphanumericData lyr = (AlphanumericData)(layers.getLayer(geoProcessingBufferPanel.getSelectedLayerName()));
        GeoProcessingPersistenceOperations.saveResultsToShapeFile(shapes, geoProcessingBufferPanel.getOutputFile());
        try {
            // TODO: Provisional. Hasta que Fernando le meta PKs al resto de ficheros.
            GeoProcessingTableOperations.saveBufferResultsToDbaseFile(geoProcessingBufferPanel.getOutputFile(), lyr, dissolveBuffer, isDbfFile);
        } catch (IncompatibleTypesException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (DriverException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (com.hardcode.gdbms.engine.data.driver.DriverException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (DriverLoadException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (NoSuchTableException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        if (!((JCheckBox)geoProcessingBufferPanel.getComponent(3)).isSelected()) ((Selectable)bufferingLayer).clearSelection();
    }
    private void mergeMethod() {
        Object[] layerNamesToMerge = geoProcessingMergePanel.getLayerSelectedNames();
        String fieldsFromLayerName = geoProcessingMergePanel.getFieldsFromLayerName();
        // TODO: Provisional. Hasta que Fernando le meta PKs al resto de ficheros.
        boolean isDbfFile = false;
        if (fieldsFromLayerName.endsWith(".shp") || fieldsFromLayerName.endsWith(".SHP")) isDbfFile = true;
        AlphanumericData lyr = (AlphanumericData)layers.getLayer(fieldsFromLayerName);
        FLayer[] layersToMerge = new FLayer[layerNamesToMerge.length];
        boolean[] selecteds = new boolean[layersToMerge.length];
        Vector allGeometriesToMerge = new Vector();
        for (int i=0;i<layerNamesToMerge.length;i++) {
            layersToMerge[i] = layers.getLayer((String)layerNamesToMerge[i]);
            if (!((Selectable)layersToMerge[i]).getSelection().isEmpty()) selecteds[i] = true;
            Geometry[] geometriesToMerge = getJtsGeometries(layersToMerge[i], selecteds[i]);
            for (int j=0;j<geometriesToMerge.length;j++) {
                allGeometriesToMerge.add(geometriesToMerge[j]);
            }
        }
        Geometry[] mergedGeometries = new Geometry[allGeometriesToMerge.size()];
        for (int i=0;i<allGeometriesToMerge.size();i++) {
            mergedGeometries[i] = (Geometry)allGeometriesToMerge.get(i);
        }
        IGeometry[] shapes = getFmapGeometries(mergedGeometries);
        AlphanumericData[] lyrs = new AlphanumericData[layersToMerge.length];
        for (int i=0;i<layersToMerge.length;i++) {
            lyrs[i] = (AlphanumericData)layersToMerge[i];
        }
        // Pasarle tb la capa con los campos.
        GeoProcessingPersistenceOperations.saveResultsToShapeFile(shapes, geoProcessingMergePanel.getOutputFile());
        try {
            GeoProcessingTableOperations.saveMergeResultsToDbaseFile(geoProcessingMergePanel.getOutputFile(), lyrs, lyr, isDbfFile);
            for (int i=0;i<lyrs.length;i++) {
                lyrs[i].getRecordset().getSelection().clear();
            }
        } catch (IncompatibleTypesException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (DriverException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (com.hardcode.gdbms.engine.data.driver.DriverException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (DriverLoadException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (NoSuchTableException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }
    private void dissolveMethod() {
        String layerToDissolveName = (String)((JComboBox)geoProcessingDissolvePanel.getComponent(2)).getSelectedItem();
        FLayer layerToDissolve = layers.getLayer(layerToDissolveName);
        boolean selected = false;
        if (((JCheckBox)geoProcessingDissolvePanel.getComponent(3)).isSelected()) selected = true;
        Geometry[] geometriesToDissolve = getJtsGeometries(layerToDissolve, selected);
        String dissolveAttributeName = (String)((JComboBox)geoProcessingDissolvePanel.getComponent(5)).getSelectedItem();
        AlphanumericData lyr = (AlphanumericData)layerToDissolve;
        // No lo puedo llevar a GeoProcessingTopologyOperations porque no es solo
        // un problema JTS. Tampoco lo puedo llevar a GeoProcessingTableOperations
        // porque tampoco es un problema solo de GDBMS. Es mixto y por el momento
        // se queda aqu.
        Vector resultGeometries = new Vector(); 
        DataSource ds;
        SelectableDataSource sds;
        try {
            ds = lyr.getRecordset();
            sds = lyr.getRecordset();
            int fieldId = ds.getFieldIndexByName(dissolveAttributeName);
            FBitSet fBitSet = sds.getSelection();
            if (fBitSet.cardinality()==0) fBitSet.set(0, (int)lyr.getRecordset().getRowCount());
            Value[] fieldValues = new Value[sds.getSelection().cardinality()];
            int k=0;
            for (int i=0;i<ds.getRowCount();i++) {
                if (fBitSet.get(i)) {
                    fieldValues[k] = ds.getFieldValue(i, fieldId);
                    k++;
                }
            }
            for (int i=0;i<fieldValues.length;i++) {
                Value valuei = fieldValues[i];
                Geometry geometryi = geometriesToDissolve[i];
                for (int j=0;j<fieldValues.length;j++) {
                    if (j!=i) {
                        Value valuej = fieldValues[j];
                        if (((BooleanValue)(valuej.equals(valuei))).getValue()) {
                            Geometry geometryj = geometriesToDissolve[j];
                                geometryi = geometryj.union(geometryi);
                        }
                    }
                }
                resultGeometries.add(geometryi);
            }
            for (int i=0;i<resultGeometries.size();i++) {
                for (int j=0;j<resultGeometries.size();j++) {
                    if (j>i) {
                        if (((Geometry)resultGeometries.get(i)).equals((Geometry)resultGeometries.get(j))) {
                            resultGeometries.remove(j);
                            j--;
                        }
                    }
                }
            }
        } catch (DriverException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (com.hardcode.gdbms.engine.data.driver.DriverException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (IncompatibleTypesException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        Geometry[] dissolvedGeometries = new Geometry[resultGeometries.size()];
        for (int i=0;i<resultGeometries.size();i++) {
            dissolvedGeometries[i] = (Geometry)resultGeometries.get(i);
        }
        //
        //Geometry[] dissolvedGeometries = GeoProcessingTopologyOperations.makeDissolve(geometriesToDissolve, lyr, dissolveAttributeName);
        IGeometry[] shapes = getFmapGeometries(dissolvedGeometries);
        // Guardo la tabla mediante gdbms. Guardo tambin las entidades JTS mediante gvSIG o usando solo gdbms (probablemente bastar con usar gdbms)
        GeoProcessingPersistenceOperations.saveResultsToShapeFile(shapes, geoProcessingDissolvePanel.getOutputFile());
        try {
            GeoProcessingTableOperations.saveDissolveResultsToDbaseFile(geoProcessingDissolvePanel.getOutputFile(), lyr, dissolveAttributeName);
            lyr.getRecordset().getSelection().clear();
        } catch (IncompatibleTypesException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (DriverException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (com.hardcode.gdbms.engine.data.driver.DriverException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (DriverLoadException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (NoSuchTableException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }
    private void spatialjoinMethod() {
        String layerToJoinName = (String)((JComboBox)geoProcessingSpatialjoinPanel.getComponent(0)).getSelectedItem();
        FLayer layerToJoin = layers.getLayer(layerToJoinName);
        String joiningLayerName = (String)((JComboBox)geoProcessingSpatialjoinPanel.getComponent(3)).getSelectedItem();
        FLayer joiningLayer = layers.getLayer(joiningLayerName);
        boolean selected = false;
        try {
            if (((JCheckBox)geoProcessingSpatialjoinPanel.getComponent(1)).isSelected()) selected = true;
            Geometry[] geometriesToJoin = getJtsGeometries(layerToJoin, selected);
            AlphanumericData lyrToJoin = (AlphanumericData)layerToJoin;
            SelectableDataSource lyrToJoinSDS = lyrToJoin.getRecordset();
            FBitSet bitsetToJoin = lyrToJoinSDS.getSelection();
            if (bitsetToJoin.cardinality()==0) bitsetToJoin.set(0, (int)lyrToJoin.getRecordset().getRowCount());
            if (((JCheckBox)geoProcessingSpatialjoinPanel.getComponent(5)).isSelected()) selected = true;
            Geometry[] joiningGeometries = getJtsGeometries(joiningLayer, selected);
            AlphanumericData joiningLyr = (AlphanumericData)joiningLayer;
            SelectableDataSource joiningLyrSDS = joiningLyr.getRecordset();
            FBitSet joiningBitset = joiningLyrSDS.getSelection();
            if (joiningBitset.cardinality()==0) joiningBitset.set(0, (int)joiningLyr.getRecordset().getRowCount());
            DataSource lyrToJoinDS;
            DataSource joiningLyrDS;
            lyrToJoinDS = lyrToJoin.getRecordset();
            joiningLyrDS = joiningLyr.getRecordset();
            // Los campos se aaden siempre. Los valores solo cuando toca ...
            String[] lyrToJoinFieldNames = new String[lyrToJoinDS.getFieldCount()];
            int[] lyrToJoinFieldTypes = new int[lyrToJoinDS.getFieldCount()];
            for (int i = 0; i < lyrToJoinDS.getFieldCount(); i++) {
                lyrToJoinFieldNames[i] = lyrToJoinDS.getFieldName(i);
                lyrToJoinFieldTypes[i] = lyrToJoinDS.getFieldType(i);
            }
            String[] joiningLyrFieldNames = new String[joiningLyrDS.getFieldCount()];
            int[] joiningLyrFieldTypes = new int[joiningLyrDS.getFieldCount()];
            for (int i = 0; i < joiningLyrDS.getFieldCount(); i++) {
                joiningLyrFieldNames[i] = joiningLyrDS.getFieldName(i);
                joiningLyrFieldTypes[i] = joiningLyrDS.getFieldType(i);
            }
            String[] joinedLyrFieldNames = new String[lyrToJoinDS.getFieldCount()+joiningLyrDS.getFieldCount()-2]; // Le quito los dos PKs
            int[] joinedLyrFieldTypes = new int[lyrToJoinDS.getFieldCount()+joiningLyrDS.getFieldCount()-2];
            int l=0;
            for (int i=0;i<lyrToJoinFieldNames.length-1;i++) {
                joinedLyrFieldNames[i] = lyrToJoinFieldNames[i];
                joinedLyrFieldTypes[i] = lyrToJoinFieldTypes[i];
                l++;
            }
            for (int i=0;i<joiningLyrFieldNames.length-1;i++) {
                joinedLyrFieldNames[i+l] = joiningLyrFieldNames[i];
                joinedLyrFieldTypes[i+l] = joiningLyrFieldTypes[i];
            }
            DataSourceFactory dsf = lyrToJoinSDS.getDataSourceFactory();
            File file = geoProcessingSpatialjoinPanel.getOutputFile();
            dsf.createFileDataSource("gdbms dbf driver", "new_table", file.getAbsolutePath().replaceAll(".shp", ".dbf"), joinedLyrFieldNames, joinedLyrFieldTypes);
            DataSource ds = dsf.createRandomDataSource("new_table");
            ds.start();
            DataWare dw = ds.getDataWare(DataSourceFactory.MANUAL_OPENING);
            dw.beginTrans();
            ds.stop();
            lyrToJoinDS.stop();
            joiningLyrDS.stop();
            lyrToJoinSDS.stop();
            joiningLyrSDS.stop();
            
            // Recorrer las geometras buscando las que intersectan. No porque
            // por ejemplo los puntos no intersectan ...
            // Ver de que combinacin se trata y calcular nearest, part of o inside
            // segn corresponda.
            // Si intersecta copiarle los atributos. O sea rellenar Values como
            // toque ...
            Value[][] values = new Value[geometriesToJoin.length][joinedLyrFieldNames.length+1]; // +1 porque faltaba el PK
            // Movida para el nearest entre puntos
            //Point[] joiningPoints = new Point[joiningGeometries.length];
            Geometry[] joiningGeometrySet = new Geometry[joiningGeometries.length];
            /*for (int i=0;i<joiningGeometries.length;i++) {
                joiningPoints[i] = (Point)joiningGeometries[i];
            }*/
            for (int i=0;i<joiningGeometries.length;i++) {
                joiningGeometrySet[i] = (Geometry)joiningGeometries[i];
            }
            //MultiPoint joiningMultiPoint = new MultiPoint(joiningPoints, new GeometryFactory());
            GeometryCollection joiningGeometryCollection = new GeometryCollection(joiningGeometrySet, new GeometryFactory());
            for (int i=0;i<geometriesToJoin.length;i++) {
                Geometry geometryToJoin = geometriesToJoin[i];
                // Movida para el nearest entre puntos
                //Coordinate[] coords = DistanceOp.closestPoints(geometryToJoin, joiningMultiPoint);
                Coordinate[] coords = DistanceOp.closestPoints(geometryToJoin, joiningGeometryCollection);
                for (int j=0;j<joiningGeometries.length;j++) {
                    Geometry joiningGeometry = joiningGeometries[j];
                    if (geometryToJoin instanceof Point && joiningGeometry instanceof Point) {
                        // Rellenar el PK. Lo tengo que hacer antes del break
                        values[i][joinedLyrFieldNames.length] = ValueFactory.createValue(i);
                        for (int k=0;k<lyrToJoinFieldNames.length-1;k++) {
                            values[i][k] = lyrToJoinDS.getFieldValue(i, k);
                        }
                        if (coords[1].equals2D(joiningGeometry.getCoordinate())) {
                            for (int k=0;k<joiningLyrFieldNames.length-1;k++) {
                                values[i][k+l] = joiningLyrDS.getFieldValue(j, k);
                            }
                            break; // En el momento encontremos un polgono que contiene a otro le metemos los valores y nos vamos
                        } else {
                            for (int k=0;k<joiningLyrFieldNames.length-1;k++) {
                                values[i][k+l] = ValueFactory.createNullValue();
                            }
                        }
                    } else if (geometryToJoin instanceof Point && joiningGeometry instanceof LineString) {
                        // part of not supported yet
                    } else if (geometryToJoin instanceof Point && joiningGeometry instanceof Polygon) {
                        // Rellenar el PK. Lo tengo que hacer antes del break
                        values[i][joinedLyrFieldNames.length] = ValueFactory.createValue(i);
                        for (int k=0;k<lyrToJoinFieldNames.length-1;k++) {
                            values[i][k] = lyrToJoinDS.getFieldValue(i, k);
                        }
                        if (joiningGeometry.contains(geometryToJoin)) {
                            for (int k=0;k<joiningLyrFieldNames.length-1;k++) {
                                System.out.println(i);
                                values[i][k+l] = joiningLyrDS.getFieldValue(j, k);
                            }
                            break; // En el momento encontremos un polgono que contiene a otro le metemos los valores y nos vamos
                        } else {
                            for (int k=0;k<joiningLyrFieldNames.length-1;k++) {
                                values[i][k+l] = ValueFactory.createNullValue();
                            }
                        }
                    } else if (geometryToJoin instanceof LineString && joiningGeometry instanceof Point) {
                        // Rellenar el PK. Lo tengo que hacer antes del break
                        values[i][joinedLyrFieldNames.length] = ValueFactory.createValue(i);
                        for (int k=0;k<lyrToJoinFieldNames.length-1;k++) {
                            values[i][k] = lyrToJoinDS.getFieldValue(i, k);
                        }
                        if (coords[1].equals2D(joiningGeometry.getCoordinate())) {
                            for (int k=0;k<joiningLyrFieldNames.length-1;k++) {
                                values[i][k+l] = joiningLyrDS.getFieldValue(j, k);
                            }
                            break; // En el momento encontremos un polgono que contiene a otro le metemos los valores y nos vamos
                        } else {
                            for (int k=0;k<joiningLyrFieldNames.length-1;k++) {
                                values[i][k+l] = ValueFactory.createNullValue();
                            }
                        }
                    } else if (geometryToJoin instanceof LineString && joiningGeometry instanceof LineString) {
                        // Rellenar el PK. Lo tengo que hacer antes del break
                        values[i][joinedLyrFieldNames.length] = ValueFactory.createValue(i);
                        for (int k=0;k<lyrToJoinFieldNames.length-1;k++) {
                            values[i][k] = lyrToJoinDS.getFieldValue(i, k);
                        }
                        //if (coords[1].equals2D(joiningGeometry.getCoordinate())) {
                        if (joiningGeometry.contains(geometryToJoin) || geometryToJoin.contains(joiningGeometry)) {
                            for (int k=0;k<joiningLyrFieldNames.length-1;k++) {
                                values[i][k+l] = joiningLyrDS.getFieldValue(j, k);
                            }
                            break; // En el momento encontremos un polgono que contiene a otro le metemos los valores y nos vamos
                        } else {
                            for (int k=0;k<joiningLyrFieldNames.length-1;k++) {
                                values[i][k+l] = ValueFactory.createNullValue();
                            }
                        }
                    } else if (geometryToJoin instanceof LineString && joiningGeometry instanceof Polygon) {
                        // Rellenar el PK. Lo tengo que hacer antes del break
                        values[i][joinedLyrFieldNames.length] = ValueFactory.createValue(i);
                        for (int k=0;k<lyrToJoinFieldNames.length-1;k++) {
                            values[i][k] = lyrToJoinDS.getFieldValue(i, k);
                        }
                        if (joiningGeometry.contains(geometryToJoin)) {
                            for (int k=0;k<joiningLyrFieldNames.length-1;k++) {
                                values[i][k+l] = joiningLyrDS.getFieldValue(j, k);
                            }
                            break; // En el momento encontremos un polgono que contiene a otro le metemos los valores y nos vamos
                        } else {
                            for (int k=0;k<joiningLyrFieldNames.length-1;k++) {
                                values[i][k+l] = ValueFactory.createNullValue();
                            }
                        }
                    } else if (geometryToJoin instanceof Polygon && joiningGeometry instanceof Point) {
                        // nearest not supported yet
                    } else if (geometryToJoin instanceof Polygon && joiningGeometry instanceof LineString) {
                        // nearest not supported yet
                    } else if (geometryToJoin instanceof Polygon && joiningGeometry instanceof Polygon) {
                        // Rellenar el PK. Lo tengo que hacer antes del break
                        values[i][joinedLyrFieldNames.length] = ValueFactory.createValue(i);
                        for (int k=0;k<lyrToJoinFieldNames.length-1;k++) {
                            values[i][k] = lyrToJoinDS.getFieldValue(i, k);
                        }
                        if (joiningGeometry.contains(geometryToJoin)) {
                            for (int k=0;k<joiningLyrFieldNames.length-1;k++) {
                                values[i][k+l] = joiningLyrDS.getFieldValue(j, k);
                            }
                            break; // En el momento encontremos un polgono que contiene a otro le metemos los valores y nos vamos
                        } else {
                            for (int k=0;k<joiningLyrFieldNames.length-1;k++) {
                                values[i][k+l] = ValueFactory.createNullValue();
                            }
                        }
                    } else {
                        // MultiPoint, LineSegment, LinearRing, MultiLineString, MultiPolygon faltan
                        // por implementar ...
                    }
                }
            }
            for (int i=0;i<lyrToJoinSDS.getSelection().cardinality();i++) {
                dw.insertFilledRow(values[i]);
            }
            /*for (int i=0;i<lyrToJoinDS.getRowCount();i++) {
                dw.insertFilledRow(values[i]);
            }*/
            dw.commitTrans();
            IGeometry[] shapes = getFmapGeometries(geometriesToJoin);
            // Guardo la tabla mediante gdbms. Guardo tambin las entidades JTS mediante gvSIG o usando solo gdbms (probablemente bastar con usar gdbms)
            // Guardar las viejas geometras.
            GeoProcessingPersistenceOperations.saveResultsToShapeFile(shapes, geoProcessingSpatialjoinPanel.getOutputFile());
            //lyrToJoinSDS.getSelection().clear();
            //joiningLyrSDS.getSelection().clear();
            lyrToJoinSDS.setSelection(new FBitSet());
            joiningLyrSDS.setSelection(new FBitSet());
        } catch (DriverException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (com.hardcode.gdbms.engine.data.driver.DriverException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (DriverLoadException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (NoSuchTableException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }
    // clip: intersect + dissolve
    /*private void clipMethod() {
        String layerToClipName = (String)((JComboBox)geoProcessingClipPanel.getComponent(0)).getSelectedItem();
        FLayer layerToClip = layers.getLayer(layerToClipName);
        boolean selected = false;
        //if (((JCheckBox)geoProcessingClipPanel.getComponent(1)).isSelected()) selected = true;
        Geometry[] geometriesToClip = getJtsGeometries(layerToClip, selected);
        String clippingLayerName = (String)((JComboBox)geoProcessingClipPanel.getComponent(3)).getSelectedItem();
        FLayer clippingLayer = layers.getLayer(clippingLayerName);
        //if (((JCheckBox)geoProcessingClipPanel.getComponent(5)).isSelected()) selected = true;
        Geometry[] clippingGeometries = getJtsGeometries(clippingLayer, selected);
        AlphanumericData lyr = (AlphanumericData)layerToClip;
        SelectableDataSource sds;
        try {
            DataSource ds;
            ds = lyr.getRecordset();
            sds = lyr.getRecordset();
            // Tener en cuenta que podemos estar trabajando solo con una seleccin
            //FBitSet fBitSet = ((Selectable)lyr).getSelection();
            String[] fieldNames = new String[sds.getFieldCount()-1];
            int[] fieldTypes = new int[sds.getFieldCount()-1];
            for (int i = 0; i < sds.getFieldCount()-1; i++) {
                fieldNames[i] = sds.getFieldName(i);
                fieldTypes[i] = sds.getFieldType(i);
            }
            DataSourceFactory dsf = sds.getDataSourceFactory();
            File file = geoProcessingClipPanel.getOutputFile();
            dsf.createFileDataSource("gdbms dbf driver", "new_table", file.getAbsolutePath().replaceAll(".shp", ".dbf"), fieldNames, fieldTypes);
            DataSource dataSource = dsf.createRandomDataSource("new_table");
            dataSource.start();
            DataWare dw = dataSource.getDataWare(DataSourceFactory.MANUAL_OPENING);
            dw.beginTrans();
            Vector values = new Vector();
            Vector geometriesClipped = new Vector();
            for (int i=0;i<geometriesToClip.length;i++) {
                Geometry geometryToClip = geometriesToClip[i];
                for (int j=0;j<clippingGeometries.length;j++) {
                    Geometry clippingGeometry = clippingGeometries[j];
                    if (clippingGeometry.intersects(geometryToClip)) {
                        geometriesClipped.add(geometryToClip.intersection(clippingGeometry));
                        Value[] row = new Value[fieldNames.length+1]; // +1 porque faltaba el PK
                        for (int m=0;m<fieldNames.length;m++) {
                            row[m] = ds.getFieldValue(i, m);
                        }
                        row[fieldNames.length] = ValueFactory.createValue(i);
                        values.add(row);
                    }
                }
            }
            // Al meter rowj.length en vez de rowj.length-1 se anula esta parte
            Vector resultGeometries = new Vector();
            for (int i=0;i<values.size();i++) {
                Value[] rowi = ((Value[])values.get(i));
                Geometry geometryi = ((Geometry)geometriesClipped.get(i));
                for (int j=0;j<values.size();j++) {
                    if (j!=i) {
                        Value[] rowj = ((Value[])values.get(j));
                        boolean areTheSameElement = true;
                        //for (int k=0;k<rowj.length;k++) {
                        for (int k=0;k<rowj.length-1;k++) {
                            Value valuei = rowi[k];
                            Value valuej = rowj[k];
                            if (!((BooleanValue)(valuej.equals(valuei))).getValue()) {
                                areTheSameElement = false;
                                break;
                            }
                        }
                        if (areTheSameElement) {
                            Geometry geometryj = ((Geometry)geometriesClipped.get(j));
                            geometryi = geometryj.union(geometryi);
                        }
                    }
                }
                resultGeometries.add(geometryi);
            }
            for (int i=0;i<resultGeometries.size();i++) {
                for (int j=0;j<resultGeometries.size();j++) {
                    if (j>i) {
                        if (((Geometry)resultGeometries.get(i)).equals((Geometry)resultGeometries.get(j))) {
                            resultGeometries.remove(j);
                            j--;
                        }
                    }
                }
            }
            for (int i=0;i<values.size();i++) {
                Value[] rowi = ((Value[])values.get(i));
                for (int j=0;j<values.size();j++) {
                    if (j>i) {
                        Value[] rowj = ((Value[])values.get(j));
                        boolean areTheSameElement = true;
                        for (int k=0;k<rowj.length-1;k++) {
                            Value valuei = rowi[k];
                            Value valuej = rowj[k];
                            if (!((BooleanValue)(valuej.equals(valuei))).getValue()) {
                                areTheSameElement = false;
                            }
                        }
                        if (areTheSameElement) {
                            values.remove(j);
                            j--;
                        }
                    }
                }
            }
            Geometry[] dissolvedGeometries = new Geometry[resultGeometries.size()];
            for (int i=0;i<resultGeometries.size();i++) {
                dissolvedGeometries[i] = (Geometry)resultGeometries.get(i);
            }
            IGeometry[] shapes = getFmapGeometries(dissolvedGeometries);
            Geometry[] clippedGeometries = new Geometry[geometriesClipped.size()];
            for (int i=0;i<geometriesClipped.size();i++) {
                clippedGeometries[i] = (Geometry)geometriesClipped.get(i);
            }
            GeoProcessingPersistenceOperations.saveResultsToShapeFile(shapes, geoProcessingClipPanel.getOutputFile());
            dataSource.stop();
            sds.stop();
            ds.stop();
            for (int i=0;i<values.size();i++) {
                dw.insertFilledRow(((Value[])values.get(i)));
            }
            dw.commitTrans();
        } catch (DriverException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (com.hardcode.gdbms.engine.data.driver.DriverException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (DriverLoadException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (NoSuchTableException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (IncompatibleTypesException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }*/
    private void clipMethod() {
        String layerToClipName = (String)((JComboBox)geoProcessingClipPanel.getComponent(0)).getSelectedItem();
        FLayer layerToClip = layers.getLayer(layerToClipName);
        boolean selected = false;
        //if (((JCheckBox)geoProcessingClipPanel.getComponent(1)).isSelected()) selected = true;
        Geometry[] geometriesToClip = getJtsGeometries(layerToClip, selected);
        String clippingLayerName = (String)((JComboBox)geoProcessingClipPanel.getComponent(3)).getSelectedItem();
        FLayer clippingLayer = layers.getLayer(clippingLayerName);
        //if (((JCheckBox)geoProcessingClipPanel.getComponent(5)).isSelected()) selected = true;
        Geometry[] clippingGeometries = getJtsGeometries(clippingLayer, selected);
        AlphanumericData lyr = (AlphanumericData)layerToClip;
        SelectableDataSource sds;
        try {
            DataSource ds;
            ds = lyr.getRecordset();
            sds = lyr.getRecordset();
            // Tener en cuenta que podemos estar trabajando solo con una seleccin
            //FBitSet fBitSet = ((Selectable)lyr).getSelection();
            String[] fieldNames = new String[sds.getFieldCount()-1];
            int[] fieldTypes = new int[sds.getFieldCount()-1];
            for (int i = 0; i < sds.getFieldCount()-1; i++) {
                fieldNames[i] = sds.getFieldName(i);
                fieldTypes[i] = sds.getFieldType(i);
            }
            DataSourceFactory dsf = sds.getDataSourceFactory();
            File file = geoProcessingClipPanel.getOutputFile();
            dsf.createFileDataSource("gdbms dbf driver", "new_table", file.getAbsolutePath().replaceAll(".shp", ".dbf"), fieldNames, fieldTypes);
            DataSource dataSource = dsf.createRandomDataSource("new_table");
            dataSource.start();
            DataWare dw = dataSource.getDataWare(DataSourceFactory.MANUAL_OPENING);
            dw.beginTrans();
            Vector values = new Vector();
            Vector geometriesClipped = new Vector();
            for (int i=0;i<geometriesToClip.length;i++) {
                Geometry geometryToClip = geometriesToClip[i];
                for (int j=0;j<clippingGeometries.length;j++) {
                    Geometry clippingGeometry = clippingGeometries[j];
                    if (clippingGeometry.intersects(geometryToClip)) {
                        geometriesClipped.add(geometryToClip.intersection(clippingGeometry));
                        Value[] row = new Value[fieldNames.length+1]; // +1 porque faltaba el PK
                        for (int m=0;m<fieldNames.length;m++) {
                            row[m] = ds.getFieldValue(i, m);
                        }
                        row[fieldNames.length] = ValueFactory.createValue(i);
                        values.add(row);
                    }
                }
            }
            // Al meter rowj.length en vez de rowj.length-1 se anula esta parte
            // o no?
            Vector resultGeometries = new Vector();
            for (int i=0;i<values.size();i++) {
                Value[] rowi = ((Value[])values.get(i));
                Geometry geometryi = ((Geometry)geometriesClipped.get(i));
                for (int j=0;j<values.size();j++) {
                    if (j!=i) {
                        Value[] rowj = ((Value[])values.get(j));
                        boolean areTheSameElement = true;
                        for (int k=0;k<rowj.length;k++) {
                        //for (int k=0;k<rowj.length-1;k++) {
                            Value valuei = rowi[k];
                            Value valuej = rowj[k];
                            if (!((BooleanValue)(valuej.equals(valuei))).getValue()) {
                                areTheSameElement = false;
                                break;
                            }
                        }
                        if (areTheSameElement) {
                            Geometry geometryj = ((Geometry)geometriesClipped.get(j));
                            geometryi = geometryj.union(geometryi);
                        }
                    }
                }
                resultGeometries.add(geometryi);
            }
            for (int i=0;i<resultGeometries.size();i++) {
                for (int j=0;j<resultGeometries.size();j++) {
                    if (j>i) {
                        if (((Geometry)resultGeometries.get(i)).equals((Geometry)resultGeometries.get(j))) {
                            resultGeometries.remove(j);
                            j--;
                        }
                    }
                }
            }
            for (int i=0;i<values.size();i++) {
                Value[] rowi = ((Value[])values.get(i));
                for (int j=0;j<values.size();j++) {
                    if (j>i) {
                        Value[] rowj = ((Value[])values.get(j));
                        boolean areTheSameElement = true;
                        for (int k=0;k<rowj.length;k++) {
                        //for (int k=0;k<rowj.length-1;k++) {
                            Value valuei = rowi[k];
                            Value valuej = rowj[k];
                            if (!((BooleanValue)(valuej.equals(valuei))).getValue()) {
                                areTheSameElement = false;
                            }
                        }
                        if (areTheSameElement) {
                            values.remove(j);
                            j--;
                        }
                    }
                }
            }
            Geometry[] dissolvedGeometries = new Geometry[resultGeometries.size()];
            for (int i=0;i<resultGeometries.size();i++) {
                dissolvedGeometries[i] = (Geometry)resultGeometries.get(i);
            }
            IGeometry[] shapes = getFmapGeometries(dissolvedGeometries);
            Geometry[] clippedGeometries = new Geometry[geometriesClipped.size()];
            for (int i=0;i<geometriesClipped.size();i++) {
                clippedGeometries[i] = (Geometry)geometriesClipped.get(i);
            }
            GeoProcessingPersistenceOperations.saveResultsToShapeFile(shapes, geoProcessingClipPanel.getOutputFile());
            dataSource.stop();
            sds.stop();
            ds.stop();
            for (int i=0;i<values.size();i++) {
                dw.insertFilledRow(((Value[])values.get(i)));
            }
            dw.commitTrans();
        } catch (DriverException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (com.hardcode.gdbms.engine.data.driver.DriverException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (DriverLoadException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (NoSuchTableException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (IncompatibleTypesException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }
    // Spatial join con Intersect
    private void intersectMethod() {
        String layerToIntersectName = (String)((JComboBox)geoProcessingIntersectPanel.getComponent(0)).getSelectedItem();
        FLayer layerToIntersect = layers.getLayer(layerToIntersectName);
        String intersectingLayerName = (String)((JComboBox)geoProcessingIntersectPanel.getComponent(3)).getSelectedItem();
        FLayer intersectingLayer = layers.getLayer(intersectingLayerName);
        boolean selected = false;
        Geometry[] geometriesToIntersect = getJtsGeometries(layerToIntersect, selected);
        Geometry[] intersectingGeometries = getJtsGeometries(intersectingLayer, selected);
        try {
            //if (((JCheckBox)geoProcessingIntersectPanel.getComponent(1)).isSelected()) selected = true;
            AlphanumericData lyrToIntersect = (AlphanumericData)layerToIntersect;
            SelectableDataSource lyrToIntersectSDS = lyrToIntersect.getRecordset();
            //if (((JCheckBox)geoProcessingIntersectPanel.getComponent(5)).isSelected()) selected = true;
            AlphanumericData intersectingLyr = (AlphanumericData)intersectingLayer;
            DataSource lyrToIntersectDS;
            DataSource intersectingLyrDS;
            lyrToIntersectDS = lyrToIntersect.getRecordset();
            intersectingLyrDS = intersectingLyr.getRecordset();
            // Los campos se aaden siempre. Los valores solo cuando toca ...
            String[] lyrToIntersectFieldNames = new String[lyrToIntersectDS.getFieldCount()];
            int[] lyrToIntersectFieldTypes = new int[lyrToIntersectDS.getFieldCount()];
            for (int i = 0; i < lyrToIntersectDS.getFieldCount(); i++) {
                lyrToIntersectFieldNames[i] = lyrToIntersectDS.getFieldName(i);
                lyrToIntersectFieldTypes[i] = lyrToIntersectDS.getFieldType(i);
            }
            String[] intersectingLyrFieldNames = new String[intersectingLyrDS.getFieldCount()];
            int[] intersectingLyrFieldTypes = new int[intersectingLyrDS.getFieldCount()];
            for (int i = 0; i < intersectingLyrDS.getFieldCount(); i++) {
                intersectingLyrFieldNames[i] = intersectingLyrDS.getFieldName(i);
                intersectingLyrFieldTypes[i] = intersectingLyrDS.getFieldType(i);
            }
            String[] intersectedLyrFieldNames = new String[lyrToIntersectDS.getFieldCount()+intersectingLyrDS.getFieldCount()-2]; // Le quito los dos PKs
            int[] intersectedLyrFieldTypes = new int[lyrToIntersectDS.getFieldCount()+intersectingLyrDS.getFieldCount()-2];
            int l=0;
            for (int i=0;i<lyrToIntersectFieldNames.length-1;i++) {
                intersectedLyrFieldNames[i] = lyrToIntersectFieldNames[i];
                intersectedLyrFieldTypes[i] = lyrToIntersectFieldTypes[i];
                l++;
            }
            for (int i=0;i<intersectingLyrFieldNames.length-1;i++) {
                intersectedLyrFieldNames[i+l] = intersectingLyrFieldNames[i];
                intersectedLyrFieldTypes[i+l] = intersectingLyrFieldTypes[i];
            }
            DataSourceFactory dsf = lyrToIntersectSDS.getDataSourceFactory();
            File file = geoProcessingIntersectPanel.getOutputFile();
            dsf.createFileDataSource("gdbms dbf driver", "new_table", file.getAbsolutePath().replaceAll(".shp", ".dbf"), intersectedLyrFieldNames, intersectedLyrFieldTypes);
            DataSource ds = dsf.createRandomDataSource("new_table");
            ds.start();
            DataWare dw = ds.getDataWare(DataSourceFactory.MANUAL_OPENING);
            dw.beginTrans();
            ds.stop();
            // Si cierro no tendr problemas a continuacin?
            lyrToIntersectSDS.stop();
            
            Vector values = new Vector();
            Vector geometriesIntersected = new Vector();
            for (int i=0;i<geometriesToIntersect.length;i++) {
                Geometry geometryToIntersect = geometriesToIntersect[i];
                String geometryToIntersectType = geometryToIntersect.getGeometryType();
                System.out.println("geometryToIntersectType = " + geometryToIntersectType);
                for (int j=0;j<intersectingGeometries.length;j++) {
                    Geometry intersectingGeometry = intersectingGeometries[j];
                    if (intersectingGeometry.intersects(geometryToIntersect)) {
                        geometriesIntersected.add(geometryToIntersect.intersection(intersectingGeometry));
                        /*for (int m=0;m<intersectedLyrFieldNames.length;m++) {
                            rows[m] = ds.getFieldValue(i, m);
                        }*/
                        Value[] rows = new Value[intersectedLyrFieldNames.length+1]; // +1 porque faltaba el PK
                        for (int k=0;k<lyrToIntersectFieldNames.length-1;k++) {
                            rows[k] = lyrToIntersectDS.getFieldValue(i, k);
                        }
                        for (int k=0;k<intersectingLyrFieldNames.length-1;k++) {
                            rows[k+l] = intersectingLyrDS.getFieldValue(j, k);
                        }
                        rows[intersectedLyrFieldNames.length] = ValueFactory.createValue(i);
                        values.add(rows);
                    }
                }
            }
            
            for (int i=0;i<geometriesIntersected.size();i++) {
                dw.insertFilledRow(((Value[])values.get(i)));
            }
            dw.commitTrans();
            Geometry[] geometriesIntersectedArray = new Geometry[geometriesIntersected.size()];
            for (int i=0;i<geometriesIntersected.size();i++) {
                geometriesIntersectedArray[i] = (Geometry)geometriesIntersected.get(i);
            }
            IGeometry[] shapes = getFmapGeometries(geometriesIntersectedArray);
            // Guardo la tabla mediante gdbms. Guardo tambin las entidades JTS mediante gvSIG o usando solo gdbms (probablemente bastar con usar gdbms)
            // Guardar las viejas geometras.
            GeoProcessingPersistenceOperations.saveResultsToShapeFile(shapes, geoProcessingIntersectPanel.getOutputFile());
        } catch (DriverException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (com.hardcode.gdbms.engine.data.driver.DriverException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (DriverLoadException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (NoSuchTableException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }
    private void unionMethod() {
        String layerToUnionName = (String)((JComboBox)geoProcessingUnionPanel.getComponent(0)).getSelectedItem();
        FLayer layerToUnion = layers.getLayer(layerToUnionName);
        boolean selected = false;
        //if (((JCheckBox)geoProcessingUnionPanel.getComponent(1)).isSelected()) selected = true;
        Geometry[] geometriesToUnion = getJtsGeometries(layerToUnion, selected);
        String uningLayerName = (String)((JComboBox)geoProcessingUnionPanel.getComponent(3)).getSelectedItem();
        FLayer uningLayer = layers.getLayer(uningLayerName);
        //if (((JCheckBox)geoProcessingUnionPanel.getComponent(5)).isSelected()) selected = true;
        Geometry[] uningGeometries = getJtsGeometries(uningLayer, selected);
        
        //if (((JCheckBox)geoProcessingIntersectPanel.getComponent(1)).isSelected()) selected = true;
        AlphanumericData lyrToUnion = (AlphanumericData)layerToUnion;
        SelectableDataSource lyrToUnionSDS;
        try {
            lyrToUnionSDS = lyrToUnion.getRecordset();
            //if (((JCheckBox)geoProcessingIntersectPanel.getComponent(5)).isSelected()) selected = true;
            AlphanumericData uningLyr = (AlphanumericData)uningLayer;
            DataSource lyrToUnionDS;
            DataSource uningLyrDS;
            lyrToUnionDS = lyrToUnion.getRecordset();
            uningLyrDS = uningLyr.getRecordset();
            // Los campos se aaden siempre. Los valores solo cuando toca ...
            String[] lyrToUnionFieldNames = new String[lyrToUnionDS.getFieldCount()];
            int[] lyrToUnionFieldTypes = new int[lyrToUnionDS.getFieldCount()];
            for (int i = 0; i < lyrToUnionDS.getFieldCount(); i++) {
                lyrToUnionFieldNames[i] = lyrToUnionDS.getFieldName(i);
                lyrToUnionFieldTypes[i] = lyrToUnionDS.getFieldType(i);
            }
            String[] uningLyrFieldNames = new String[uningLyrDS.getFieldCount()];
            int[] uningLyrFieldTypes = new int[uningLyrDS.getFieldCount()];
            for (int i = 0; i < uningLyrDS.getFieldCount(); i++) {
                uningLyrFieldNames[i] = uningLyrDS.getFieldName(i);
                uningLyrFieldTypes[i] = uningLyrDS.getFieldType(i);
            }
            String[] unedLyrFieldNames = new String[lyrToUnionDS.getFieldCount()+uningLyrDS.getFieldCount()-2]; // Le quito los dos PKs
            int[] unedLyrFieldTypes = new int[lyrToUnionDS.getFieldCount()+uningLyrDS.getFieldCount()-2];
            int l=0;
            for (int i=0;i<lyrToUnionFieldNames.length-1;i++) {
                unedLyrFieldNames[i] = lyrToUnionFieldNames[i];
                unedLyrFieldTypes[i] = lyrToUnionFieldTypes[i];
                l++;
            }
            for (int i=0;i<uningLyrFieldNames.length-1;i++) {
                unedLyrFieldNames[i+l] = uningLyrFieldNames[i];
                unedLyrFieldTypes[i+l] = uningLyrFieldTypes[i];
            }
            DataSourceFactory dsf = lyrToUnionSDS.getDataSourceFactory();
            File file = geoProcessingUnionPanel.getOutputFile();
            dsf.createFileDataSource("gdbms dbf driver", "new_table", file.getAbsolutePath().replaceAll(".shp", ".dbf"), unedLyrFieldNames, unedLyrFieldTypes);
            DataSource ds = dsf.createRandomDataSource("new_table");
            ds.start();
            DataWare dw = ds.getDataWare(DataSourceFactory.MANUAL_OPENING);
            dw.beginTrans();
            ds.stop();
            lyrToUnionSDS.stop();
            
            Vector values = new Vector();
            
            //Geometry[] unedGeometries = GeoProcessingTopologyOperations.makeUnion(geometriesToUnion, uningGeometries);
            Geometry[] result = new Geometry[geometriesToUnion.length];
            Vector geometriesUned = new Vector();
            //Selectable selectable = (Selectable)layerToUnion;
            //FBitSet fBitSet = new FBitSet();
            // Metememos en geometriesUned las intersecciones.
            for (int i=0;i<geometriesToUnion.length;i++) {
                Geometry geometryToUnion = geometriesToUnion[i];
                for (int j=0;j<uningGeometries.length;j++) {
                    Geometry uningGeometry = uningGeometries[j];
                    if (uningGeometry.intersects(geometryToUnion)) {
                        //fBitSet.set(i);
                        geometriesUned.add(geometryToUnion.intersection(uningGeometry));
                        Value[] intersectRow = new Value[unedLyrFieldNames.length+1];
                        for (int k=0;k<lyrToUnionFieldNames.length-1;k++) {
                            intersectRow[k] = lyrToUnionDS.getFieldValue(i, k);
                        }
                        for (int k=0;k<uningLyrFieldNames.length-1;k++) {
                            intersectRow[k+l] = uningLyrDS.getFieldValue(j, k);
                        }
                        intersectRow[unedLyrFieldNames.length] = ValueFactory.createValue(i);
                        values.add(intersectRow);
                        //geometriesUned.add(geometryToUnion.difference(uningGeometry));
                    }
                }
            }
            // Rellenamos geometriesUned2 con estas intersecciones.
            Vector geometriesUned2 = new Vector();
            for (int i=0;i<geometriesUned.size();i++) {
                geometriesUned2.add(geometriesUned.get(i));
            }
            //int num = geometriesUned.size();
            // Aado a geometriesUned2 las geometras de geometriesToUnion que no
            // intersectan con uningGeometries.
            for (int i=0;i<geometriesToUnion.length;i++) {
                Geometry geometryToUnion = geometriesToUnion[i];
                boolean intersectsWithAnyGeometry = false;
                for (int j=0;j<geometriesUned.size();j++) {
                    Geometry geometryUned = (Geometry)geometriesUned.get(j);
                    if (geometryToUnion.intersects(geometryUned)) intersectsWithAnyGeometry = true;
                }
                if (intersectsWithAnyGeometry) {
                    //Geometry difference = geometryToUnion.difference(geometryUned);
                    //geometriesUned2.add(difference);
                } else {
                    geometriesUned2.add(geometryToUnion);
                    Value[] unionRow = new Value[unedLyrFieldNames.length+1];
                    for (int k=0;k<lyrToUnionFieldNames.length-1;k++) {
                        unionRow[k] = lyrToUnionDS.getFieldValue(i, k);
                    }
                    for (int k=0;k<uningLyrFieldNames.length-1;k++) {
                        unionRow[k+l] = ValueFactory.createNullValue();
                    }
                    unionRow[unedLyrFieldNames.length] = ValueFactory.createValue(i);
                    values.add(unionRow);
                }
            }
            // Aado a geometriesUned2 las geometras de uningGeometries que no
            // intersectan con geometriesToUnion.
            for (int i=0;i<uningGeometries.length;i++) {
                Geometry uningGeometry = uningGeometries[i];
                boolean intersectsWithAnyGeometry = false;
                for (int j=0;j<geometriesUned.size();j++) {
                    Geometry geometryUned = (Geometry)geometriesUned.get(j);
                    if (uningGeometry.intersects(geometryUned)) intersectsWithAnyGeometry = true;
                }
                if (intersectsWithAnyGeometry) {
                    //Geometry difference = jtsUningGeometry.difference(jtsGeometryUned);
                    //jtsGeometriesUned2.add(difference);
                } else {
                    geometriesUned2.add(uningGeometry);
                    Value[] uningRow = new Value[unedLyrFieldNames.length+1];
                    for (int k=0;k<lyrToUnionFieldNames.length-1;k++) {
                        uningRow[k] = ValueFactory.createNullValue();
                    }
                    for (int k=0;k<uningLyrFieldNames.length-1;k++) {
                        uningRow[k+l] = uningLyrDS.getFieldValue(i, k);
                    }
                    uningRow[unedLyrFieldNames.length] = ValueFactory.createValue(i);
                    values.add(uningRow);
                }
            }
            //System.out.println("geometriesUned2.size() = " + geometriesUned2.size());
            // Le quito los cachos que ya tengo a las geometrias que faltan de
            // uningGeometries.
            BitSet bitSet1 = new BitSet();
            bitSet1.clear();
            //int k1=0;
            for (int i=0;i<geometriesUned2.size();i++) {
                Geometry geometryUned2 = (Geometry)geometriesUned2.get(i);
                for (int j=0;j<uningGeometries.length;j++) {
                    //System.out.println("!(uningGeometries[j].equals(geometryUned2)) = " + !(uningGeometries[j].equals(geometryUned2)));
                    if (uningGeometries[j].intersects(geometryUned2) && !(uningGeometries[j].equals(geometryUned2))) {
                        uningGeometries[j] = uningGeometries[j].difference(geometryUned2);
                        //bitSet1.set(k1);
                        bitSet1.set(j);
                    }
                    //k1++;
                }
            }
            // Le quito los cachos que ya tengo a las geometrias que faltan de
            // geometriesToUnion.
            BitSet bitSet2 = new BitSet();
            bitSet2.clear();
            int k2=0;
            for (int i=0;i<geometriesUned2.size();i++) {
                Geometry geometryUned2 = (Geometry)geometriesUned2.get(i);
                for (int j=0;j<geometriesToUnion.length;j++) {
                    //System.out.println("!(geometriesToUnion[j].equals(geometryUned2)) = " + !(geometriesToUnion[j].equals(geometryUned2)));
                    if (geometriesToUnion[j].intersects(geometryUned2) && !(geometriesToUnion[j].equals(geometryUned2))) {
                        geometriesToUnion[j] = geometriesToUnion[j].difference(geometryUned2);
                        bitSet2.set(k2);
                        //bitSet2.set(j);
                    }
                    k2++;
                }
            }
            //for (int i=0;i<bitSet2.size();i++) {
            //    System.out.println("bitSet2.get(" + i + ") = " + bitSet2.get(i));
            //}
            // Ahora que les he quitado los cachos que ya tena, le meto los features que
            // faltaban de uningGeometries
            for (int i=0;i<uningGeometries.length;i++) {
                if (bitSet1.get(i)) {
                    geometriesUned2.add(uningGeometries[i]);
                    Value[] uningRow = new Value[unedLyrFieldNames.length+1];
                    for (int k=0;k<lyrToUnionFieldNames.length-1;k++) {
                        uningRow[k] = ValueFactory.createNullValue();
                    }
                    for (int k=0;k<uningLyrFieldNames.length-1;k++) {
                        uningRow[k+l] = uningLyrDS.getFieldValue(i, k);
                    }
                    uningRow[unedLyrFieldNames.length] = ValueFactory.createValue(i);
                    values.add(uningRow);
                }
            }
            // Ahora que les he quitado los cachos que ya tena, le meto los features que
            // faltaban de geometriesToUnion
            for (int i=0;i<geometriesToUnion.length;i++) {
                if (bitSet2.get(i)) {
                    geometriesUned2.add(geometriesToUnion[i]);
                    Value[] unionRow = new Value[unedLyrFieldNames.length+1];
                    for (int k=0;k<lyrToUnionFieldNames.length-1;k++) {
                        unionRow[k] = lyrToUnionDS.getFieldValue(i, k);
                    }
                    for (int k=0;k<uningLyrFieldNames.length-1;k++) {
                        unionRow[k+l] = ValueFactory.createNullValue();
                    }
                    unionRow[unedLyrFieldNames.length] = ValueFactory.createValue(i);
                    values.add(unionRow);
                }
            }
            //selectable.setSelection(fBitSet);
            result = new Geometry[geometriesUned2.size()];
            for (int i=0;i<geometriesUned2.size();i++) {
                result[i] = (Geometry)geometriesUned2.get(i);
            }
            
            for (int i=0;i<result.length;i++) {
                dw.insertFilledRow(((Value[])values.get(i)));
            }
            dw.commitTrans();
            
            IGeometry[] shapes = getFmapGeometries(result);
            // Guardo la tabla mediante gdbms. Guardo tambin las entidades JTS mediante gvSIG o usando solo gdbms (probablemente bastar con usar gdbms)
            GeoProcessingPersistenceOperations.saveResultsToShapeFile(shapes, geoProcessingUnionPanel.getOutputFile());
        } catch (DriverException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (com.hardcode.gdbms.engine.data.driver.DriverException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (DriverLoadException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (NoSuchTableException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }
    public double[] getBufferDistancesFromTable(FLayer bufferingLayer) {
        AlphanumericData lyr = (AlphanumericData)bufferingLayer;
        DataSource ds;
        double[] bufferDistances = null;
        try {
            ds = lyr.getRecordset();
            int fieldId = ds.getFieldIndexByName(getSelectedField());
            Value[] fieldValues = new Value[(int)ds.getRowCount()];
            SelectableDataSource selectableDataSource = (SelectableDataSource)ds;
            FBitSet fBitSet = selectableDataSource.getSelection();
            if (fBitSet.cardinality()==0) {
                fBitSet.set(0, (int)lyr.getRecordset().getRowCount());
            }
            Value[] valuesSelected = new Value[fBitSet.cardinality()];
            int j = 0;
            for (int i=0;i<ds.getRowCount();i++) {
                boolean isSelected = fBitSet.get(i);
                fieldValues[i] = ds.getFieldValue(i, fieldId);
                if (isSelected) {
                    valuesSelected[j] = fieldValues[i];
                    j++;
                }
            }
            bufferDistances = new double[fBitSet.cardinality()];
            for (int i=0;i<fBitSet.cardinality();i++) {
                bufferDistances[i] = new Double(valuesSelected[i].toString()).doubleValue();
            }
        } catch (DriverException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (com.hardcode.gdbms.engine.data.driver.DriverException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        return bufferDistances;
    }
    private String getSelectedField() {
        return ((String)((JComboBox)geoProcessingBufferPanel.getComponent(6)).getSelectedItem());
    }
    private int getQuadrantSegments() {
        return (Integer.valueOf((String)((JComboBox)geoProcessingBufferPanel.getComponent(11)).getSelectedItem())).intValue();
    }
    private void closeMethod() {
        if (PluginServices.getMainFrame() == null)
            ((JDialog) (getParent().getParent().getParent().getParent())).dispose();
        else
            PluginServices.getMDIManager().closeView(GeoProcessingPanel.this);
    }
   }  //  @jve:decl-index=0:visual-constraint="10,10"
